@@ -1,28 +1,15 @@
-from dataclasses import dataclass, field
+import copy
 import functools
 import logging
 import re
-import copy
-import inspect
-from typing import (
-    Any,
-    Iterable,
-    List,
-    Sequence,
-    Set,
-    Tuple,
-    Type,
-    Union,
-    Dict,
-    Optional,
-    TypeVar,
-    Generic,
-    NamedTuple,
-    get_type_hints
-)
+from dataclasses import dataclass, field
+from inspect import getmembers, isclass, isroutine
+from typing import (Any, Dict, Generic, Iterable, List, NamedTuple, Optional,
+                    Sequence, Set, Tuple, Type, TypeVar, Union, get_type_hints)
+
 import yaml
 
-from .utils import hybridmethod, MISSING
+from .utils import MISSING, hybridmethod
 
 
 # get logger for this file, mostly used for debugging
@@ -30,16 +17,18 @@ LOGGER = logging.getLogger(__name__)
 
 
 # Type declaration for classes
+T = TypeVar("T")
 CN = TypeVar('CN', bound='ConfigNode')
 CV = TypeVar('CV', bound='ConfigPlaceholderVar')
 CP = TypeVar('CP', bound='ConfigParam')
 CR = TypeVar('CR', bound='ConfigNodeProps')
 CE = TypeVar('CE', bound='ConfigRegistryReference')
-
+PS = TypeVar('PS', bound='ParameterSpec')
 
 # Constants
 CONFIG_NODE_PROPERTIES_NAME = '__node__'
 CONFIG_NODE_REGISTRY_SEPARATOR = '@'
+CONFIG_PARAM_CLI_FORMAT = r'([a-zA-Z_]\w*)(\.([a-zA-Z_]\w*))*=.*'
 CONFIG_PLACEHOLDER_VAR_TEMPLATE = (
     r'\$\{(([a-zA-Z_]\w*)\.?)*(%s[a-zA-Z_]+)*\}' %
     CONFIG_NODE_REGISTRY_SEPARATOR
@@ -47,12 +36,29 @@ CONFIG_PLACEHOLDER_VAR_TEMPLATE = (
 
 
 class ParameterSpec(NamedTuple):
+    """Holds a parameter spec. Can be instantiated from either string or
+    by passing the attributes directly."""
+
     name: str
-    default: Any
-    type: Type[Any]
+    default: T
+    type: Type[T]
+
+    @classmethod
+    def from_string(cls: Type[PS], string: str) -> PS:
+        if not re.findall(CONFIG_PARAM_CLI_FORMAT, string):
+            msg = f"Cannot parse string '{string}' into a parameter"
+            raise ValueError(msg)
+        name, default = string.split('=', 1)
+        default = yaml.safe_load(default)
+        return cls(name=name, default=default, type=type(default))
+
+    def to_dict(self: PS) -> dict:
+        outdict = None
+        for part in self.name.split('.')[::-1]:
+            outdict = {part: outdict} if outdict else {part: self.default}
+        return outdict
 
 
-T = TypeVar("T")
 class ConfigParam(Generic[CP]):
     """Type wrapper to indicate parameters in a config node."""
     type: T
@@ -66,7 +72,7 @@ class ConfigParam(Generic[CP]):
 class ConfigRegistryReference(Generic[CE]):
     """Extract registry references and resolve them"""
     def __init__(
-        self,
+        self: CE,
         param_name: str = None,
         registry_ref: str = None,
         registry_args: Sequence[Any] = None
@@ -75,10 +81,10 @@ class ConfigRegistryReference(Generic[CE]):
         self.registry_ref = registry_ref
         self.registry_args = registry_args or []
 
-    def name_as_placeholder_variable(self):
+    def name_as_placeholder_variable(self: CE):
         """Split the name (if provided) to be used
         as the path for placeholder variable"""
-        if self.param_name is not None:
+        if self.param_name:
             return self.param_name.split('.')
         return None
 
@@ -94,7 +100,10 @@ class ConfigRegistryReference(Generic[CE]):
                        registry_ref=registry_ref,
                        registry_args=args)
 
-    def resolve(self, param_value: Any, nodes_cls: Sequence[CN] = None, **kwargs) -> Any:
+    def resolve(self: CE,
+                *args,
+                nodes_cls: Sequence[CN] = None,
+                **kwargs) -> Any:
         """Resolves and instantiates a reference to a registry object using
         `param_value`. `nodes_cls` can be a list of ConfigNode objects to
         merge with no override to the registry reference (if the registry
@@ -103,7 +112,7 @@ class ConfigRegistryReference(Generic[CE]):
 
         if self.registry_ref is None:
             # no-op if there is no registry ref!
-            return param_value
+            return args[0] if len(args) == 1 else args
 
         nodes_cls = nodes_cls or []
 
@@ -111,26 +120,28 @@ class ConfigRegistryReference(Generic[CE]):
         registry_reference = ConfigRegistry.get(self.registry_ref)
 
         for node_cls in nodes_cls:
-            if inspect.isclass(node_cls) and not issubclass(node_cls, ConfigNode):
+
+            if isclass(node_cls) and not issubclass(node_cls, ConfigNode):
                 msg = ('The registry reference resolver has receive an object'
                        f'of type {node_cls}, which is not a ConfigNode')
                 raise ValueError(msg)
-            if (inspect.isclass(registry_reference) and
+
+            if (isclass(registry_reference) and
                 not issubclass(registry_reference, ConfigNode)):
-                msg = ('The registry reference resolver has received one or more'
-                       'ConfigNode, but they cannot be merged with a registry '
-                       f'reference of type {type(registry_reference)}')
+                msg = ('The registry reference resolver has received one or '
+                       'more ConfigNode, but they cannot be merged with a '
+                       f'registry reference of type {type(registry_reference)}')
                 raise ValueError(msg)
+
             registry_reference = registry_reference >> node_cls
 
-        return registry_reference(param_value, *self.registry_args, **kwargs)
+        return registry_reference(*args, *self.registry_args, **kwargs)
 
     @classmethod
     def contains(cls, string: str) -> bool:
         return CONFIG_NODE_REGISTRY_SEPARATOR in string
 
 
-
 class ConfigNodeProps(Generic[CR]):
     def __init__(self: CR, node: CN, name: str, parent: CN = None):
         self.node = node
@@ -163,11 +174,14 @@ class ConfigNodeProps(Generic[CR]):
     def set_name(self: CR, node_name: Union[str, None]):
             # set the full path to this node as its name
         self.short_name = self.cls_name if node_name is None else node_name
-        self.long_name = (self.short_name if self.is_root() else
-                          f'{self.get_props(self.parent).long_name}.{self.short_name}')
+        self.long_name = (
+            self.short_name if self.is_root() else
+            f'{self.get_props(self.parent).long_name}.{self.short_name}'
+        )
 
     @hybridmethod
-    def get_annotations(cls: Type[CR], node_cls: Type[CN]) -> Dict[str, ConfigParam]:
+    def get_annotations(cls: Type[CR],
+                        node_cls: Type[CN]) -> Dict[str, ConfigParam]:
         # these are all annotations for parameters for this node; we use
         # them to cast param values to the right type as we parse a config
         annotations = {name: annotation for name, annotation in
@@ -186,7 +200,7 @@ class ConfigNodeProps(Generic[CR]):
 
         # We use getmembers instead of __dict__ because it
         # resolves inheritance (__dict__ does not!)
-        all_members = dict(inspect.getmembers(node_cls))
+        all_members = dict(getmembers(node_cls))
 
         defaults = {name: all_members[name]
                     for name in cls.get_annotations(node_cls)
@@ -212,9 +226,9 @@ class ConfigNodeProps(Generic[CR]):
 
         # We use getmembers instead of __dict__ because it
         # resolves inheritance (__dict__ does not!)
-        all_members = dict(inspect.getmembers(node_cls))
+        all_members = dict(getmembers(node_cls))
         subnodes = {name: cls_ for name, cls_ in all_members.items()
-                    if inspect.isclass(cls_) and issubclass(cls_, ConfigNode)}
+                    if isclass(cls_) and issubclass(cls_, ConfigNode)}
         return subnodes
 
     @get_subnodes.instancemethod
@@ -222,8 +236,9 @@ class ConfigNodeProps(Generic[CR]):
         return type(self).get_subnodes(self.node_cls)
 
     @hybridmethod
-    def get_all_cls_members(cls: Type[CR], node_cls: Type[CN]) -> Dict[str, Any]:
-        all_non_routines = inspect.getmembers(node_cls, lambda a: not(inspect.isroutine(a)))
+    def get_all_cls_members(cls: Type[CR],
+                            node_cls: Type[CN]) -> Dict[str, Any]:
+        all_non_routines = getmembers(node_cls, lambda a: not(isroutine(a)))
         return {name: value for name, value in all_non_routines
                 if not((name.startswith('__') and name.endswith('__')) or
                         name == '_is_protocol')}
@@ -233,7 +248,8 @@ class ConfigNodeProps(Generic[CR]):
         return type(self).get_all_cls_members(self.node_cls)
 
     @hybridmethod
-    def get_all_parameters(cls: Type[CR], node_cls: Type[CN]) -> Dict[str, ParameterSpec]:
+    def get_all_parameters(cls: Type[CR],
+                           node_cls: Type[CN]) -> Sequence[ParameterSpec]:
         """Get all parameters for this node as well as for its subnodes."""
 
         all_parameters = []
@@ -242,7 +258,6 @@ class ConfigNodeProps(Generic[CR]):
         defaults = cls.get_defaults(node_cls)
 
         for param_name, param_annotation in annotations.items():
-
             all_parameters.append(
                 ParameterSpec(name=param_name,
                               type=param_annotation.type,
@@ -256,21 +271,28 @@ class ConfigNodeProps(Generic[CR]):
                                   type=param_spec.type,
                                   default=param_spec.default)
                 )
-
         return all_parameters
 
     @get_all_parameters.instancemethod
-    def get_all_parameters(self: CR) -> Dict[str, ParameterSpec]:
+    def get_all_parameters(self: CR) -> Sequence[ParameterSpec]:
         return type(self).get_all_parameters(self.node_cls)
 
-    def assign_param(self, name: str, value: Any):
+    def assign_param(self, name: str, value: Any, annotation: Any = None):
         # printing some debug info
-        LOGGER.debug(f'[ASSIGN VAR][{self.long_name}][{self.cls_name}] {name}={value}')
+        LOGGER.debug(f'[ASSIGN VAR][{self.long_name}]'
+                     f'[{self.cls_name}] {name}={value}')
 
         # do the actual assignment
         self.param_keys.add(name)
         setattr(self.node, name, value)
 
+        # optionally set up annotation if provided
+        # (for more advanced uses)
+        if annotation is not None:
+            if not hasattr(self.node, '__annotations__'):
+                setattr(self.node, '__annotations__', {})
+            setattr(self.node.__annotations__, name, annotation)
+
     def get_params_names(self: CR) -> Sequence[str]:
         return tuple(self.param_keys)
 
@@ -288,12 +310,12 @@ class ConfigNodeProps(Generic[CR]):
         else:
             KeyError(f'`{key}` is not a parameter in {self.long_name}')
 
-
     def is_root(self: CR) -> bool:
         return self.parent == self.node
 
     def get_root(self: CR) -> CN:
-        """Traverse the configuration this node is part of to find the root node"""
+        """Traverse the configuration this node is part of to
+        find the root node"""
         if self.is_root():
             return self.node
         else:
@@ -323,11 +345,14 @@ class ConfigNodeProps(Generic[CR]):
             raise RuntimeError(msg)
 
     def to_dict(self: CR) -> Dict[str, Any]:
-        return {k: (self.get_props(v).to_dict() if isinstance(v, ConfigNode)
-                    else (repr(v) if isinstance(v, ConfigPlaceholderVar) else v))
+        return {k: (self.get_props(v).to_dict()
+                    if isinstance(v, ConfigNode) else
+                    (repr(v) if isinstance(v, ConfigPlaceholderVar) else v))
                 for k, v in self.node}
 
-    def to_yaml(self: CR, *args: Sequence[Any], **kwargs: Dict[str, Any]) -> str:
+    def to_yaml(self: CR,
+                *args: Sequence[Any],
+                **kwargs: Dict[str, Any]) -> str:
         return yaml.safe_dump(self.to_dict(), *args, **kwargs)
 
     @hybridmethod
@@ -402,11 +427,14 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
         ) -> None:
         # Parsing comes in 5 phases, each one is in a separate code block!
 
-        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PHASE 0 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PHASE 0 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         # This is mostly preliminaries
 
-        # Create a property object for this node; the object will bind itself to it
-        node_props = ConfigNodeProps(node=self, name=__name__, parent=__parent__)
+        # Create a property object for this node; the object will bind
+        # itself to it
+        node_props = ConfigNodeProps(node=self,
+                                     name=__name__,
+                                     parent=__parent__)
 
         # get annnotations, defaults, and subnodes. Will be used to look up the
         # right types for values, get their default value, and instantiate any
@@ -428,9 +456,9 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
             msg = (f'Config to `{node_props.cls_name}` should be dict, '
                    f'but received {type(config)} instead!')
             raise ValueError(msg)
-        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
-        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PHASE 1 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PHASE 1 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         # Override with options that are provided in `config` dictionary
         for param_name, param_value in config.items():
             # a little lambda function to log how we are doing
@@ -441,7 +469,8 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
             if ConfigRegistryReference.contains(param_name):
                 # we have a registry reference! we need to extract the
                 # reference and get the config from the registry.
-                registry_reference = ConfigRegistryReference.from_str(param_name)
+                registry_reference = ConfigRegistryReference.\
+                    from_str(param_name)
                 param_name = registry_reference.param_name
             else:
                 registry_reference = None
@@ -453,7 +482,8 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
                 registry_reference is None):
                 debug_call('0.not_supported')
                 # CASE 0: we can't add this key to this config; raise an error.
-                msg = f'Parameter "{param_name}" not supported in "{node_props.cls_name}"'
+                msg = (f'Parameter "{param_name}" not '
+                       f'supported in "{node_props.cls_name}"')
                 raise KeyError(msg)
 
             if registry_reference is not None:
@@ -514,7 +544,7 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
                 # parameter is, so we first extract the type.
                 param_type = annotations[param_name].type
 
-                if (inspect.isclass(param_type) and
+                if (isclass(param_type) and
                     issubclass(param_type, ConfigNode)):
                     # CASE 3.1: sometimes the type of a paramemter is
                     #           ConfigNode or a subclass of it, making
@@ -522,9 +552,16 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
                     #           the constructor instead of just `param_value`.
                     #           We accomplish that with a partial decorator.
                     debug_call('3.1.ann_subnode')
-                    param_type = functools.partial(
-                        param_type, __parent__=self, __name__=param_name
-                    )
+
+                    # It only makes sense to do this replacement
+                    # if the parameter provided is a dict or a ConfigNode,
+                    # because those can be parsed by a config node.
+                    if isinstance(param_value, (dict, ConfigNode)):
+                        param_type = functools.partial(
+                            param_type, __parent__=self, __name__=param_name
+                        )
+                    else:
+                        param_type = lambda x: x
                 elif isinstance(param_value, param_type):
                     # We cast to param_type, but only if we absolutely
                     # have to. This prevents unwanted initializations is
@@ -549,13 +586,27 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
                 #         the subnode and set that as parameter value
                 debug_call('4.subnode')
 
-                param_value = subnodes[param_name](config=param_value,
-                                                   __parent__=self,
-                                                   __name__=param_name)
+                if isinstance(param_value, (dict, ConfigNode)):
+                    # note that calling the subnode init method here
+                    # only make sense if we are dealing with a supported
+                    # type (a dict or another config node). If not, it
+                    # does not make much sense to call the constructor,
+                    # as it would lead to errors down the line.
+                    # NOTE: we could call it if param_value is none too,
+                    # but we assume that if a user has specified a None
+                    # here, they do not want to instantiate a subnode.
+                    # if they still wish to, they could simply pass {} instead.
+                    debug_call('4.1.subnode_applied')
+                    param_value = subnodes[param_name](config=param_value,
+                                                       __parent__=self,
+                                                       __name__=param_name)
+                else:
+                    debug_call('4.2.subnode_skipped')
             elif __flex_node__:
-                # CASE 5: I don't recognize this key, but I'm in flex config mode
-                #         so I'll just add it to this node object (config_value
-                #         is already asssigned, that's why we just pass here).
+                # CASE 5: I don't recognize this key, but I'm in flex config
+                #         mode so I'll just add it to this node object
+                #         (config_value is already asssigned, that's why we
+                #         just pass here).
                 debug_call('5.unrecognized')
             else:
                 # CASE ∞: Something went wrong and you reached a technically
@@ -566,21 +617,24 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
 
             # add parameter to set of parameters in this node and assign it
             node_props.assign_param(name=param_name, value=param_value)
-        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
-        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PHASE 2 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PHASE 2 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         # Try using default values for parameters with no override. To find
         # which parameters are still missing, we get a tuple of parameter names
         # initialized so far from the node properties object; then we use it
         # to remove parameter names from the set of all annotations.
-        missing_params = set(annotations).difference(set(node_props.get_params_names()))
+        missing_params = set(annotations).difference(
+            set(node_props.get_params_names())
+        )
 
         for param_name in missing_params:
             if param_name not in defaults:
                 # This parameter was not overwritten by the configuration
                 # provided to this __init__ method; however, the paramer
                 # doesn't have a default, so we need to raise an error.
-                msg = f'parameter "{param_name}" is missing for "{node_props.cls_name}"'
+                msg = (f'parameter "{param_name}" is '
+                       f'missing for "{node_props.cls_name}"')
                 raise ValueError(msg)
 
             # We got lucky! We found a default value for this annotated
@@ -608,32 +662,36 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
             # add parameter to set of parameters in this node and
             # assign its DEFAULT VALUE
             node_props.assign_param(name=param_name, value=param_value)
-        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
-        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PHASE 3 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PHASE 3 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         # Last set of assignments! Like we did with annotations, we need to
-        # check if there are one or more submodules that have not been initialized
-        # already through the config provided by the user.
-        missing_subnodes = set(subnodes).difference(set(node_props.get_params_names()))
+        # check if there are one or more submodules that have not been
+        # initialized already through the config provided by the user.
+        missing_subnodes = set(subnodes).difference(
+            set(node_props.get_params_names())
+        )
 
         for param_name in missing_subnodes:
             # initialize the subnode here
-            param_value = subnodes[param_name](__parent__=self, __name__=param_name)
+            param_value = subnodes[param_name](__parent__=self,
+                                               __name__=param_name)
 
             # add the subnode to the config
             node_props.assign_param(name=param_name, value=param_value)
-        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
-        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PHASE 4 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PHASE 4 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         # Finally, if this is a root class, we want to take care of replacing
         # variables with actual values; not going to do any late binding a la
         # hydra, with some minimal logic to catch cyclical assignments
         if node_props.is_root():
             node_props.apply_vars()
-        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
     def __iter__(self: CN) -> Iterable[Tuple[str, Any]]:
-        """Get a iterable of names and parameters in this node, including subnodes"""
+        """Get a iterable of names and parameters in this node,
+        including subnodes"""
         yield from ((k, self[k]) for k in
                     ConfigNodeProps.get_props(self).get_params_names())
 
@@ -650,7 +708,8 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
         if key in node_props.get_params_names():
             return getattr(self, key)
         else:
-            raise KeyError(f'`{key}` is not a parameter in {node_props.long_name}')
+            msg = f'`{key}` is not a parameter in {node_props.long_name}'
+            raise KeyError(msg)
 
     def __len__(self: CN) -> int:
         return sum(1 for _ in self)
@@ -676,7 +735,9 @@ class ConfigNode(Generic[CN], metaclass=MetaConfigNode):
 class VarMatch:
     path: Sequence[str]
     match: str
-    registry: ConfigRegistryReference = field(default_factory=ConfigRegistryReference)
+    registry: ConfigRegistryReference = field(
+        default_factory=ConfigRegistryReference
+    )
 
 
 class ConfigPlaceholderVar(Generic[CV]):
@@ -760,6 +821,22 @@ class ConfigPlaceholderVar(Generic[CV]):
             if isinstance(placeholder_substitution, ConfigNode):
                 ConfigNodeProps.get_props(placeholder_substitution).apply_vars()
 
+            # trick of the century: sometimes, by doing variable resolution,
+            # we end up with another variable! in that case, we simply
+            # tell the parent node with the class to do variable resolution too.
+            if isinstance(placeholder_substitution, ConfigPlaceholderVar):
+                # asking the parent to apply all its vars!
+                ConfigNodeProps.get_props(
+                    placeholder_substitution.parent_node).apply_vars()
+
+                # note that even if we apply the substitution,
+                # the value inside `placeholder_substitution` placeholder
+                # var does not change, so we need to manually fish it out
+                placeholder_substitution = getattr(
+                    placeholder_substitution.parent_node,
+                    placeholder_substitution.param_name
+                )
+
             # we make a copy of the object to avoid unwanted side effects
             placeholder_substitution = copy.deepcopy(placeholder_substitution)
 
@@ -798,6 +875,8 @@ class ConfigPlaceholderVar(Generic[CV]):
             # type to cast if necessary.
             replaced_value = self.param_config_type(replaced_value)
 
+        # finally done with variable resolution, let's set the
+        # parent to this value.
         setattr(self.parent_node, self.param_name, replaced_value)
 
     @classmethod
@@ -811,10 +890,56 @@ class ConfigPlaceholderVar(Generic[CV]):
 class ConfigFlexNode(ConfigNode):
     """Just like a ConfigNode, except it allows for additional
     parameters other than the ones provided with annotations"""
-    def __init__(
-        self: CN,
-        *args: List[Any],
-         __flex_node__: bool = True,
-        **kwargs: Dict[str, Any]
-    ) -> None:
+    def __init__(self: CN,
+                 *args: List[Any],
+                 __flex_node__: bool = True,
+                 **kwargs: Dict[str, Any]) -> None:
         super().__init__(*args, __flex_node__=True, **kwargs)
+
+
+class DictOfConfigNodes:
+    """A special type of node that contains a dictionary of ConfigNodes.
+    Useful for when you want to provide a bunch of nodes, but you are
+    not sure what the name of keys are. Usage:
+
+    ```python
+    from espresso_config import (
+        NodeConfig, DictOfConfigNodes, ConfigParam
+    )
+
+    class ConfigA(NodeConfig):
+        p: ConfigParam(int)
+
+    class RootConfig(NodeConfig):
+        dict_of_configs: ConfigParam(DictOfConfigNodes(ConfigA)) = {}
+
+    ```
+
+    and in the corresponding yaml file:
+
+    ```yaml
+    dict_of_configs:
+        first_config:
+            p: 1
+        second_config:
+            p: 2
+        ...
+    ```
+    """
+
+    class _Wrapper:
+
+        def __new__(cls, config=None, *args, **kwargs):
+            parsed = {}
+
+            for k, v in (config or {}).items():
+                node = super().__new__(cls)
+                node.__init__(v, *args, **kwargs)
+                parsed[k] = node
+
+            flex_node_wrapper_cls = type(cls.__name__, (ConfigFlexNode, ), {})
+            return flex_node_wrapper_cls(parsed, *args, **kwargs)
+
+    def __new__(cls, node_cls: ConfigFlexNode) -> ConfigFlexNode:
+        return type(f'DictOf{node_cls.__name__}',
+                    (cls._Wrapper, node_cls), {})
